/*
    PROCESSOR: ATMega328p @ 8MHz
        Fuses    Low: 0xe2;  High: 0xd1;  Ext: 0x06;
                            -U lfuse:w:0xe2:m -U hfuse:w:0xd1:m -U efuse:w:0xfe:m
                            8Mhz internal osci.; 1.8V brownout; preserve EEPROM

        todo:
                in random mode, count nr of cycles. That way you can use a lower duration, while still allow longer animations (blockmove) to cycle at least twice
                fix animations that don't set every single led. e.g. going to blockmove, will look weird if comming from the previous animation
                X reset frame when animations changes
                X change animation on frame==0 ?
                X when the uC starts in standby, pressing the button will wake up, but also change to the next mode (while it should only wake up)
                X accellerometer
                X don't go to standby after WDT reset

        done:
                measure battery voltage
                grayscale
                brownout detection
                WDT



        BUTTON CHANGE SETTING:
                single-press                : next mode
                press-and-hold              : standby (+ previous mode)
                double-press                : next speed/color
                double-press-and-hold       : next speed/color every 1s
                triple-press                : random-mode

        PINOUT
                PB0  LEFT-LED-1
                PB1  LEFT-LED-2
                PB2  LEFT-LED-3
                PB3  LEFT-LED-4 MOSI (OC2A)
                PB4  LEFT-LED-5 MISO
                PB5  SCK
                PB6  x
                PB7  x

                PC0  x
                PC1  STATUS-LED
                PC2  CE (enable boot converter)
                PC3  ACCEL-VCC
                PC4  SDA
                PC5  SCL
                PC6  RESET

                PD0   (RX)
                PD1   (TX)
                PD2  BUTTON (INT0)
                PD3  RIGHT-LED-1 (OC2B)
                PD4  RIGHT-LED-2
                PD5  RIGHT-LED-3
                PD6  RIGHT-LED-4
                PD7  RIGHT-LED-5

                ADC7 VBAT
*/

/*
        INCLUDES
*/
#include <avr/sleep.h>
#include <avr/power.h>
#include <avr/wdt.h>
#include <avr/eeprom.h>
#include <avr/delay.h>

//todo: disable for productions
//#define DEBUG_ENABLED

#ifdef DEBUG_ENABLED
#define print(a) (Serial.print(a))
#else
#define print(a) 
#endif


/*
     CONSTANTS
*/
const uint8_t BUTTON_PIN = 2;
const uint8_t LED_PIN = A1;
const uint8_t CE_PIN = A2;

const byte ANIMATIONS_COUNT = 20;
const byte RANDOM_ANIMATIONS_COUNT = ANIMATIONS_COUNT - 5; // don't use last 5 animations in random-mode
const byte SPEEDS_COUNT = 5;
const byte MAX_BRIGHTNESS = 4;
const byte LED_COUNT = 15;

void (*animations[ANIMATIONS_COUNT])(); //array of all the animation-functions

/*
     GLOBAL VARIABLES
*/
volatile boolean standby = false;
volatile boolean preStandby = false; // set to true while the button is still being pressed, before going to sleep
volatile boolean randomAnimation = true;// start with random
volatile uint8_t animationIndex = 0;
volatile uint8_t speeds[ANIMATIONS_COUNT] = {2,};// 0..4
uint16_t frame = 0;

uint8_t leds[2][LED_COUNT] = {{0,}, {0,}};



/*
     SETUP
*/
void setup()
{
    wdt_reset();
#ifdef DEBUG_ENABLED
    Serial.begin(57600);
#endif
    print("setup()");


    /*//led pattern 00000101000001010000 (draws 33uA)
    PORTD = _BV(PORTD6);
    DDRD = 0B10111000;
    PORTB = _BV(PORTB3);
    DDRB = 0B00010111; */
    /*    
    //led pattern 010000010000010
    PORTD = 0B10001000;
    DDRD  = 0B01010000;
    PORTB = 0B00010001;
    DDRB  = 0B00001010;
    delay(600);
    */

    // Enable watchdog
    wdt_reset();
    wdt_enable(WDTO_120MS);


    //randomAnimation = false;//todo: change when enough animations

    uint8_t i = 0;
    animations[i++] = randomFlash;
    animations[i++] = eyes;
    animations[i++] = closeEyes;
    animations[i++] = bursts;
    animations[i++] = equalizer;
    animations[i++] = larsonScanner;
    animations[i++] = fill;
    animations[i++] = fill2;
    animations[i++] = blockMove;
    animations[i++] = blockFill;
    animations[i++] = chase;
    animations[i++] = chase2;
    animations[i++] = chase3;
    animations[i++] = wave;
    animations[i++] = fade;
    animations[i++] = fadeLeftRight;
    animations[i++] = alternate;
    animations[i++] = alternateLeftRight;
    animations[i++] = allOn;
    animations[i++] = allLow;
    // last option is random animation
    print(i);
    print(" number of animations\n");
    if( i!=ANIMATIONS_COUNT )
        print("Incorrect animation count\n");

    // default speed is 2 (of 0..4)
    for (byte i = 0; i < sizeof(speeds); i++)
        speeds[i] = 2;

/*
 * bettery to have the unit on, than go off during an event
    standby = true; // default to OFF, to prevent accidentally turning ON after e.g. a short
    if ( MCUSR & _BV(EXTRF) ) //todo: detect WDT reset
    {
        // detect reset (so I don't have to turn the unit on after programming it) (note: power on reset flag PORF is set after reprogramming)
        standby = false;
        MCUSR = 0;
    }
*/    


    pinMode(BUTTON_PIN, INPUT_PULLUP);
    EIMSK = _BV(INT0);// enable INT0 interrupt
    EICRA = _BV(ISC01);// falling edge

    pinMode(LED_PIN, OUTPUT);

    // enable boost converter (leaving this an input, will sometimes cause the boost converter to switch of (it only has a 10 MOhm pull-up))
    digitalWrite(CE_PIN, HIGH);
    pinMode(CE_PIN, OUTPUT);

    // debug pins for logic analyzer
    pinMode(A0, OUTPUT);
    pinMode(A3, OUTPUT);


    randomSeed(analogRead(A6));

    showBatteryVoltage();//note: a delay will be added in `loop`; to show this battery indicator


    // Save power
    ADCSRA &= ~_BV(ADEN);  // disable ADC
    ACSR |= _BV(ACD);     //disable the analog comparator (note: the analog comparator is used to trigger interrupts based on the values on the analog pins)
    // Power Reduction Register
    power_adc_disable();      // ADC converter
    power_twi_disable();      // TWI (I2C)
    power_spi_disable();      // SPI
    power_timer2_disable();   // Timer 2
#ifndef DEBUG_ENABLED
    power_usart0_disable();   // Serial (USART)
#endif
    //power_timer0_disable();   // Timer 0
    //power_timer1_disable();   // Timer 1

    /*
            PORTB = 0x00;// all LOW
            DDRB  = 0xFF;// all outputs
            PORTC = 0x00;// all LOW
            DDRC  = _BV(DDC4) | _BV(DDC2) | _BV(DDC1) | _BV(DDC0);// unconnected set to output
            PORTD = _BV(PORTD2);// all LOW, except INT0 (button)
            DDRD  = _BV(DDD6) | _BV(DDD5) | _BV(DDD3) | _BV(DDD1);// some set to output
    */

    // Setup timer 1 (16-bit timer)
    // fast-pwm mode, TOP: ICR1
    TCCR1A = _BV(WGM11);
    TCCR1B = _BV(WGM13) | _BV(WGM12);
    TCCR1B |= _BV(CS11);// CK/8 prescaler => 1000kHz
    ICR1 = 50 - 1; //;// 8000000/8/50=20kHz    /16 leds =>  1.25Hz    /8 grayscale => 156Hz

// super slowmotion
//TCCR1B |= _BV(CS12) | _BV(CS12) ;// CK/1024
//ICR1 = 65536/32;  

    TIMSK1 |= _BV(ICIE1);

}


/*
     MAIN LOOP
*/
void loop()
{
    unsigned long startTime;
    unsigned long lastAnimationChange = millis()-11000;

    // wait a bit, to show the battery voltage indicator
    for( byte i=0; i<80; i++ )
    {
        delay(10);
        wdt_reset();
    }


    while ( true )
    {
        startTime = millis();

        // turn off all leds while button is being helt (prior to going to standy). This will visually indicate that the button has been pressed long enough to turn the device off.
        while ( preStandby && !standby )
        {
            allOff();
            wdt_reset();
        }

        // sleep the µC
        while ( standby )
        {
            allOff();
            sleep();
            standby = false;
            setBlinks(1);
        }

        // check if a new random animation needs to be selected
        if ( randomAnimation )
        {
            if ( (millis() - lastAnimationChange > 11000 && frame == 1) || millis() - lastAnimationChange > 22000  ) // change to different mode every 11 seconds (11 instead of 10, because blockmove take 10.9ms, so using 11 will make it cycle twice
            {
                setRandomAnimation();
                lastAnimationChange = millis();
            }
        }

        // Run currently selected animation
        ///PORTC |= _BV(PORTC3);//for bebugging timing (ISR takes 34µs out of 257µs per cycle)
        (*animations[animationIndex])();
        ///PORTC &= ~_BV(PORTC3);// for debugging timing

        // add delay, depending on the selected speed
        static const int speedDelays[SPEEDS_COUNT] = {15, 10, 5, 3, 2};//{100, 66, 45, 30, 20};// every speed is around 1.5 times faster

        while ( millis() - startTime < speedDelays[speeds[animationIndex]] ) {};
        frame++;

        wdt_reset();

    }
}



void setRandomAnimation()
{
    static uint8_t shuffledList[RANDOM_ANIMATIONS_COUNT];
    static uint8_t shuffleIndex = 254;
    static const uint8_t MIN_DISTANCE = 4;

    if ( shuffleIndex == 254 )
    {
        // initialize; fill with 0, 1, 2, etc.
        for (uint8_t i = 0; i < RANDOM_ANIMATIONS_COUNT; i++)
            shuffledList[i] = i;
    }


    shuffleIndex++;
    if ( shuffleIndex >= RANDOM_ANIMATIONS_COUNT - (MIN_DISTANCE - 1) )
    {
        // restart
        shuffleIndex = 0;

        // shuffle everything that has already been done
        // Fisher–Yates_shuffle
        for (uint8_t i = 0; i < RANDOM_ANIMATIONS_COUNT - (MIN_DISTANCE - 1) - 1; i++)
        {
            uint8_t j = random(i, RANDOM_ANIMATIONS_COUNT - (MIN_DISTANCE - 1));
            uint8_t t = shuffledList[i];
            // swap [i] with [j]
            shuffledList[i] = shuffledList[j];
            shuffledList[j] = t;
        }

        // move last items (that haven't been done yet) to the front
        for (uint8_t i = 0; i < (MIN_DISTANCE - 1); i++)
        {
            uint8_t j = RANDOM_ANIMATIONS_COUNT - (MIN_DISTANCE - 1) + i;
            uint8_t t = shuffledList[i];
            shuffledList[i] = shuffledList[j];
            shuffledList[j] = t;
        }

    }
    animationIndex = shuffledList[shuffleIndex];
    frame = -1;
}


void showBatteryVoltage()
{
    // 3V = full (15 leds); 2.2V = empty (1 led)
    int a = analogRead(7);//ADC7
    int mv = 5000L * a / 1024;
    if( mv<2200 )
        mv = 2200;
    byte barSize = (long)(mv-2200)*(LED_COUNT-1)/(3000-2200)+1;

    for( byte i=0; i<LED_COUNT; i++ )
        setLed(i, 0);
    for( byte i=0; i<barSize && i<LED_COUNT; i++ )
        setLed(i, 1);
}

/*
        ANIMATIONS
*/
static inline void randomFlash()
{
    if( (frame & 0x7) != 0 )
        return;
    frame = 0;

    byte index = emptyLedIndex();
    byte index2 = emptyLedIndex();

    for( byte i=0; i< LED_COUNT; i++ )
    {
        byte b = getLed(i);
        if( b>0 )
            setLed(i, b-1);
    }

    setLed(index, MAX_BRIGHTNESS);
    setLed(index2, MAX_BRIGHTNESS);
}


static inline void eyes()
{
    static const byte pupilSize = 3;
    static int8_t position = 2; // position 0 means far left; -9 means just out of view on the left
    static int8_t targetPosition = 3;

    if ( frame % 8 != 0 )
        return;

    if( position!=targetPosition )
    {
        if( position<targetPosition )
            position++;
        else
            position--;
    }
    else
    {
        frame = 0;
        if( random(10)==0 )
            targetPosition = random(0, LED_COUNT-pupilSize)-(MAX_BRIGHTNESS-1);//todo: offset from left, and from right
    }

    // blank leds on the left
    for( int8_t i=0; i<position; i++ )
        setLedUnmirrored(i, 0);
    
    int8_t i = position;
    // left side of iris; increase brigtness
    for( int8_t j=0; j<MAX_BRIGHTNESS-1; j++ )
    {
        i++;
        if( (i>=0) && (i<LED_COUNT) )
            setLedUnmirrored(i, j+1);
    }
    // pupil
    for( int8_t j=0; j<pupilSize; j++ )
    {
        i++;
        if( (i>=0) && (i<LED_COUNT) )
            setLedUnmirrored(i, MAX_BRIGHTNESS);
    }
    // right side of iris; decrease brigtness
    for( int8_t j=0; j<MAX_BRIGHTNESS-1; j++ )
    {
        i++;
        if( (i>=0) && (i<LED_COUNT) )
            setLedUnmirrored(i, MAX_BRIGHTNESS-1-j);
    }

    // blank leds on the right
    for( int8_t j=i; j<LED_COUNT; j++ )
    {
        if( j>=0 )
            setLedUnmirrored(j, 0);
    }
}


static inline void closeEyes()
{
    uint16_t index = frame / 8;
    if ( index > (LED_COUNT/2 + MAX_BRIGHTNESS)*2 )
    {
        frame = 0;
        index = 0;
    }

    if( index < LED_COUNT/2 + MAX_BRIGHTNESS )
    {
        byte center = (LED_COUNT+1)/2 - 1;
        for( byte i=0; i<(LED_COUNT+1)/2; i++ )
        {
            byte b = 0;
            if( i+index>center )
                b = min(MAX_BRIGHTNESS, i+index-center);
            setLed(i, b);
            setLed(LED_COUNT-1-i, b);
        }
    }
    else
    {
        index -= LED_COUNT/2 + MAX_BRIGHTNESS;
        for( byte i=0; i<(LED_COUNT+1)/2; i++ )
        {
            byte b = MAX_BRIGHTNESS;
            if( index>i )
                if( index-i<MAX_BRIGHTNESS )
                    b = MAX_BRIGHTNESS+i-index;
                else
                    b=0;
            setLed(i, b);
            setLed(LED_COUNT-1-i, b);
        }
    }
}


static inline void bursts()
{
    if( frame < 9 )
        return;
    frame = 0;

    bool clear = getLed(LED_COUNT-1)==0;

    for( byte i=0; i<LED_COUNT; i++ )
    {
        byte b = getLed(i);
        if( b==MAX_BRIGHTNESS && (i>0) )
            setLed(i-1, MAX_BRIGHTNESS );
        if( b!=0 )
            setLed(i, b-1);
    }

    if( clear && random(5)==0 )
        setLed(LED_COUNT-1, MAX_BRIGHTNESS);
}


static inline void equalizer()
{    
    static byte target = 6;
    static byte lastTarget = 0;

    bool done = true;
    byte i;
    if( target>lastTarget )
    {
        for( i=0; i<=target; i++ )
        {
            byte b = getLed(i);
            if( b!=MAX_BRIGHTNESS )
            {
                setLed(i, b+1);
                done = false;
                break;
            }
        }
    }
    else
    {
        for( byte i=LED_COUNT-1; i>target; i--)
        {
            byte b = getLed(i);
            if( b!=0 )
            {
                setLed(i, b-1);
                done = false;
                break;
            }            
        }
    }
    
    if( done )
    {
        lastTarget = target;
        target = random(0, LED_COUNT);
        frame = 0;
    }


}


static inline void larsonScanner()
{
    uint8_t index = frame / 4;
    if ( index >= 2 * LED_COUNT )
    {
        frame = 0;
        index = 0;
    }

    if ( index < LED_COUNT )
    {
        // move forth
        for ( byte i = 0; i < LED_COUNT; i++ )
        {
            if( index>MAX_BRIGHTNESS && i < index-MAX_BRIGHTNESS )
                    setLed(i, 0);
            else if ( i<=index )
                setLed(i, MAX_BRIGHTNESS+i-index);
            else
                setLed(i, 0);
        }
    }
    else
    {
        // move back
        index = LED_COUNT * 2 - 1 - index;
        for ( byte i = 0; i < LED_COUNT; i++ )
        {
            if( i<index )
                    setLed(i, 0);
            else if ( i<index+MAX_BRIGHTNESS )
                setLed(i, MAX_BRIGHTNESS+index-i);
            else
                setLed(i, 0);
        }
    }
}
//todo: larson scanner with different speed, left and right eye


static inline void fill() // light up leds, one by one, from left to right; then turn of leds, one by one, from left to right
{
    uint8_t index = frame / 4;
    if ( index >= LED_COUNT * 2 )
    {
        frame = 0;
        index = 0;
    }

    if ( index < LED_COUNT )
    {
        for ( byte i = 0; i <= index; i++ )
            setLed(i, MAX_BRIGHTNESS);
    }
    else
    {
        for ( byte i = 0; i <= index - LED_COUNT; i++ )
            setLed(i, 0);
    }
}


static inline void fill2() // light up leds, one by one, from left to right; then turn of leds, one by one, from left to right; and then in reverse
{
    uint8_t index = frame / 4;
    if ( index >= LED_COUNT * 4 )
    {
        frame = 0;
        index = 0;
    }

    if ( index < LED_COUNT )
    {
        for ( byte i = 0; i <= index; i++ )
            setLed(i, MAX_BRIGHTNESS);
    }
    else if ( index < LED_COUNT * 2 )
    {
        for ( byte i = 0; i <= index - LED_COUNT; i++ )
            setLed(i, 0);
    }
    if ( index < LED_COUNT * 3 )
    {
        for ( byte i = 0; i <= index - LED_COUNT * 2; i++ )
            setLed(LED_COUNT - 1 - i, MAX_BRIGHTNESS);
    }
    else
    {
        for ( byte i = 0; i <= index - LED_COUNT * 3; i++ )
            setLed(LED_COUNT - 1 - i, 0);
    }
}


static inline void blockFill()
{
    uint16_t index = frame / 4;
    byte i;

    if ( index >= (LED_COUNT * (LED_COUNT + 1) / 2) * 2 - LED_COUNT ) // sum of 1..n = n(n+1)/2   -2 because the last 20 steps of unfill are identical to the first 20 steps of fill, which looks weird
    {
        frame = 0;
        index = 0;
    }

    if ( index < (LED_COUNT * (LED_COUNT + 1)) / 2 )
    {
        // Fill
        for ( i = LED_COUNT - 1; i > 0; i-- )
        {
            if ( index > i )
            {
                index -= i + 1;
            }
            else
                break;
        }
        for ( byte j = 0; j < index; j++ )
            setLed(j, 0);// clear all leds before the moving lit led
        setLed(index, MAX_BRIGHTNESS);// moving led
        for ( byte j = index + 1; j < i; j++ ) // clear leds after moving lit led// you don't really need to clear the leds between the moving led, and the lit leds, since they are already clear from the previous frame
            setLed(j, 0);
        for ( byte j = i; j < LED_COUNT; j++ )
            setLed(j, MAX_BRIGHTNESS);// light up all leds that have been moved

    }
    else
    {
        // Un-fill
        index -= (LED_COUNT * (LED_COUNT + 1)) / 2 - 1; // -1 for skip the first step, which is identical to the last step of the fill
        for ( i = 0; i < LED_COUNT; i++ )
        {
            if ( index > i )
            {
                index -= i + 1;
            }
            else
                break;
        }
        for ( byte j = 0; j < i + 1; j++ ) // todo i+1 could be bigger than 20 (I think)
            setLed(LED_COUNT - 1 - j, 0);
        for ( byte j = i + 1; j < LED_COUNT; j++ )
            setLed(LED_COUNT - 1 - j, MAX_BRIGHTNESS);
        setLed(LED_COUNT - 1 - (i - index), MAX_BRIGHTNESS);
        //todo: first and 20th led are incorrect
    }
}


static inline void blockMove()
{
    uint16_t index = frame / 4;
    byte i;
    bool reverse = false;
    if ( index >= (LED_COUNT * LED_COUNT) * 2 )
    {
        frame = 0;
        index = 0;
    }

    if ( index == 0 )
        for ( byte i = 0; i < LED_COUNT; i++ )
            setLed(i, 0, MAX_BRIGHTNESS);

    if ( index >= LED_COUNT * LED_COUNT )
    {
        index -= LED_COUNT * LED_COUNT;
        reverse = true;
    }
    byte done = index / LED_COUNT;
    byte remainder = index % LED_COUNT;

    byte onLedIndex = LED_COUNT - done + remainder;
    byte offLedIndex = onLedIndex - 1;

    if ( reverse )
    {
        onLedIndex  = LED_COUNT * 2 - 1 - onLedIndex;
        offLedIndex = LED_COUNT * 2 - 1 - offLedIndex;
    }

    if ( onLedIndex < LED_COUNT )
        setRightLed(onLedIndex, MAX_BRIGHTNESS);
    else
        setLeftLed(LED_COUNT * 2 - 1 - onLedIndex, MAX_BRIGHTNESS);

    if ( offLedIndex < LED_COUNT )
        setRightLed(offLedIndex, 0);
    else
        setLeftLed(LED_COUNT * 2 - 1 - offLedIndex, 0);
}


static inline void chase() //chase
{
    if ( frame / 8 >= 8 )
        frame = 0;

    // 4 leds on (and 4 leds off), moving inward
    for ( byte i = 0; i < LED_COUNT; i++ )
        if ( (frame / 8 + i) / 4 % 2 == 0 )
            setLed(i, MAX_BRIGHTNESS);
        else
            setLed(i, 0);
}


static inline void chase2()
{
    if ( frame / 8 >= 9 )
        frame = 0;

    // 3 leds on (and 6 leds off), moving inward
    for ( byte i = 0; i < LED_COUNT; i++ )
        if ( (frame / 8 + i) % 9 < 3 )
            setLed(i, MAX_BRIGHTNESS);
        else
            setLed(i, 0);
}


static inline void chase3()
{
    if ( frame / 8 >= 9 )
        frame = 0;

    // 3 leds on (and 6 leds off), moving left- to right
    for ( byte i = 0; i < LED_COUNT; i++ )
    {
        byte brightness = (frame / 8 + i) % 9 < 3 ? MAX_BRIGHTNESS : 0;
        setLedUnmirrored(i, brightness);
    }
}


static inline void wave()
{
    if ( frame / 16 >= 8 )
        frame = 0;

    for ( byte i = 0; i < LED_COUNT; i++ )
    {
        byte brightness = (frame / 16 + i) % 8 + 1;      // 1 2 3 4 5 6 7 8
        if ( brightness > MAX_BRIGHTNESS )              //         5 6 7 8
            brightness = MAX_BRIGHTNESS * 2 + 1 - brightness; //         4 3 2 1  (9-x)
        setLed(i, brightness);
    }
}


static inline void fade()
{
    uint16_t index = frame / 16;
    if ( index > MAX_BRIGHTNESS * 2 )
    {
        frame = 0;
        index = 0;
    }
    byte brightness;
    if ( index < MAX_BRIGHTNESS )
        brightness = index;
    else
        brightness = MAX_BRIGHTNESS * 2 - index;

    for ( byte i = 0; i < LED_COUNT; i++ )
        setLed(i, brightness);
}


static inline void fadeLeftRight()
{
    uint16_t index = frame / 16;
    if ( index > MAX_BRIGHTNESS * 2 )
    {
        frame = 0;
        index = 0;
    }
    byte brightness;
    if ( index < MAX_BRIGHTNESS )
        brightness = index;
    else
        brightness = MAX_BRIGHTNESS * 2 - index;

    for ( byte i = 0; i < LED_COUNT; i++ )
    {
        setLeftLed(i, brightness);
        setRightLed(i, MAX_BRIGHTNESS-brightness);
    }
}


static inline void alternate()
{
    if ( frame >= 100 )
        frame = 0;
    for ( byte i = 0; i < LED_COUNT; i++ )
        setLed(i, frame < 50 ? MAX_BRIGHTNESS : 0);
}


static inline void alternateLeftRight()
{
    if ( frame >= 100 )
        frame = 0;
    if ( frame < 50 )
    {
        for ( byte i = 0; i < LED_COUNT; i++ )
            setLed(i, 0, MAX_BRIGHTNESS);
    }
    else if ( frame < 100 )
    {
        for ( byte i = 0; i < LED_COUNT; i++ )
            setLed(i, MAX_BRIGHTNESS, 0);
    }
}


static inline void allOn()
{
    frame = 0;
    for (byte i = 0; i < LED_COUNT; i++)
        setLed(i, MAX_BRIGHTNESS);
}


static inline void allLow()
{
    frame = 0;
    for (byte i = 0; i < LED_COUNT; i++)
        setLed(i, 1);
}


static inline void allOff()
{
    frame = 0;
    for (byte i = 0; i < LED_COUNT; i++)
        setLed(i, 0);
}


static inline void deadlock()
{
    while(1);
}




void printLeds()
{
    for( byte i=0; i<LED_COUNT; i++ )
    {
        print(getLed(i));
    }
    print("\n");
}


static inline uint8_t emptyLedIndex()
{
    byte index = random(0, LED_COUNT);
    for( byte i=0; i<LED_COUNT; i++)
    {
        if( getLed(index)!=0 )
                index = (index+11) % LED_COUNT;
        else
                break;
    }
    return index;
}







static inline void handleButton()
{
    // assumes it get's called every 4ms (200 times per second)

    // todo:
    //          use timer to sample button
    //       triple press doesn't always work
    //
    //       + skip 100 cycle waiting period on release, if button was previously held.
    //       press and hold after standby doesn't put it back to sleep (but kind of an edge case)
    //       + don't apply action after waking from standby
    //       + don't apply action on release. if action is already done in the press-and-hold.
    //           Otherwise, after releasing from a press-and-hold the action will be performed 1 additional time.
    //       + why doesn't the button-release wake up the uC from standby (after long single press)
    //       +   it only does so if you time it right


    static const uint16_t historyMask = 0x0000f;
    static uint16_t pinHistory;
    static bool pressed = false;// pressed or not
    static bool hold = false;   // press-and-hold
    static bool ignoreNextPress = false;
    static uint8_t pressCount = 0;// number of consecutive presses
    static uint16_t stateDuration = 0;// time the button is being pressed/released

    stateDuration++;

    bool changed = false;

    bool pinState = digitalRead(BUTTON_PIN);
    pinHistory = (pinHistory << 1) | pinState;
    if ( !pressed && (pinHistory & historyMask) == 0 ) //i.e. xxxxxxxxxxxx0000
    {
        // pressed
        pressed = true;
        changed = true;
        stateDuration = 0;
        pressCount++;
        setBlinks(1);
    }
    else if ( pressed && (pinHistory & historyMask) == historyMask ) //i.e. xxxxxxxxxxxx1111
    {
        // released
        pressed = false;
        changed = true;
        stateDuration = 0;
    }

    if ( pressed && ( (!hold && stateDuration > (750 / 4)) || (hold && stateDuration > (1500 / 4)) ) ) // press and hold (hold after 750ms, repeat after 1.5s)
    {
        hold = true;
        if ( pressCount == 1 )
        {
            // standby
            preStandby = true;// turn of the light, but wait till release  to go to sleep
            setBlinks(-1);//light led to indicate going to standby (maybe disable other leds)
        }
        else if ( pressCount == 2 )
        {
            // change speed
            nextSpeed();
            setBlinks(2);
        }
        else if ( pressCount == 3 )
        {
            // change brightness
            //nextBrightness();
            setBlinks(3);
        }
        stateDuration = 0;
    }
    else if ( !pressed && pressCount > 0 && (stateDuration > (500 / 4) || hold) ) // release (without re-press)
    {
        if ( ignoreNextPress )
        {
            ignoreNextPress = false;
            preStandby = false;
            setBlinks(0);
        }
        else if ( hold )
        {
            // don't apply action on release, if action is already done in the press-and-hold
            hold = false;
            if ( pressCount == 1 )
            {
                setBlinks(0);
                //digitalWrite(LED_PIN, LOW);
                ignoreNextPress = true;
                standby = true;
                preStandby = false;
            }
        }
        else // regular release (without hold)
        {
            if ( pressCount == 1 )
            {
                // next mode
                nextMode();
                //setBlinks(1);
            }
            else if ( pressCount == 2 )
            {
                // change speed
                nextSpeed();
                //setBlinks(2);
            }
            else if ( pressCount == 3 )
            {
                randomAnimation = true;
                // change brightness
                //nextBrightness();
                //setBlinks(3);
            }
        }
        pressCount = 0;
    }
    blink();
    //digitalWrite(13, pressed);

}

static inline nextMode()
{
    if ( randomAnimation )
    {
        randomAnimation = false;
        animationIndex = 0; // start back at zero
    }
    else
    {
        animationIndex++;
        if ( animationIndex >= ANIMATIONS_COUNT )
        {
            randomAnimation = true;
            animationIndex = 0;
        }
    }
}

static inline nextSpeed()
{
    speeds[animationIndex] = (speeds[animationIndex] + 1) % SPEEDS_COUNT;
}


const uint8_t BLINK_ON_DURATION = (25 / 4);
const uint8_t BLINK_OFF_DURATION = (250 / 4);
int16_t blinkCounter = 0;

void setBlinks(int8_t count)
{
    // blink 20ms on 20ms off
    if ( count == 0 )
        blinkCounter = 0;
    else if ( count == -1 )
        blinkCounter = -1;
    else
        blinkCounter = count * (BLINK_ON_DURATION + BLINK_OFF_DURATION) - BLINK_OFF_DURATION;
}

void blink()
{
    if ( blinkCounter == 0 )
    {
        digitalWrite(LED_PIN, LOW);
    }
    else if ( blinkCounter == -1 )
    {
        digitalWrite(LED_PIN, HIGH);
    }
    else
    {
        blinkCounter--;// decrease counter before if-condition, because I use `<`. Otherwise the first iteration the led will be off, and the last iteration the led won't light (last blink will otherwise be shorter) (all verified using excel)
        if ( blinkCounter % (BLINK_ON_DURATION + BLINK_OFF_DURATION) < BLINK_ON_DURATION )
            digitalWrite(LED_PIN, HIGH);
        else
            digitalWrite(LED_PIN, LOW);
    }
}



inline static void sleep()
{

    // turn of timers
    uint8_t tccr1a = TCCR1A;
    uint8_t tccr1b = TCCR1B;
    TCCR1A = 0x00;
    TCCR1B = 0x00;

    // turn of status led
    digitalWrite(LED_PIN, LOW);

    // set all led pins to LOW and INPUT
    PORTB = 0;
    DDRB = 0;
    PORTD = _BV(PORTD2);
    DDRD = 0;

    //todo: turn off accelerometer

    //todo: turn off boost converter (maybe)

    // disable watchdog
    wdt_disable();

    // sleep uC
    cli();
    set_sleep_mode(SLEEP_MODE_PWR_DOWN);
    sleep_enable();
    sleep_bod_disable();
    sei();
    sleep_cpu();
    sleep_disable();

    // Enable watchdog
    wdt_enable(WDTO_120MS);

    // turn timers back on
    TCCR1A = tccr1a;
    TCCR1B = tccr1b;
}



/*inline static void oe()
    {
        //active low
        PORTB &= ~_BV(PORTB2); // pin 10 low
    } */

uint8_t getLed(uint8_t index)
{
    if ( index < LED_COUNT )
    {
        return leds[0][index];
    }
}

void setLed(uint8_t index, uint8_t value)
{
    if ( index < LED_COUNT )
    {
        leds[0][index] = value;
        leds[1][index] = value;
    }
    else
        print("setLed out of bounds\n");
}

void setLed(uint8_t index, uint8_t leftValue, uint8_t rightValue)
{
    if ( index < LED_COUNT )
    {
        leds[0][index] = leftValue;
        leds[1][index] = rightValue;
    }
    else
        print("setLed2 out of bounds\n");
}

void setLeftLed(uint8_t index, uint8_t value)
{
    if ( index < LED_COUNT )
    {
        leds[0][index] = value;
    }
    else
        print("setLeftLed out of bounds\n");
}

void setRightLed(uint8_t index, uint8_t value)
{
    if ( index < LED_COUNT )
    {
        leds[1][index] = value;
    }
    else
        print("setRightLed out of bounds\n");
}

void setLedInverse(uint8_t index, uint8_t value)
{
    if ( index < LED_COUNT )
    {
        leds[0][LED_COUNT - 1 - index] = value;
        leds[1][LED_COUNT - 1 - index] = value;
    }
    else
        print("setLed out of bounds\n");
}

void setLedUnmirrored(uint8_t index, uint8_t value)
{
    if ( index < LED_COUNT )
    {
        leds[0][index] = value;
        leds[1][LED_COUNT - 1 - index] = value;
    }
    else
        print("setLedUnmirrored out of bounds\n");
}



static const uint8_t ledDdrB[20]  = {0b00000011, 0b00000011, 0b00000110, 0b00000110, 0b00001100, 0b00001100, 0b00011000, 0b00011000, 0b00010100, 0b00010100, 0b00000101, 0b00000101, 0b00001001, 0b00001001, 0b00001010, 0b00001010, 0b00010010, 0b00010010, 0b00010001, 0b00010001};
static const uint8_t ledPortB[20] = {0b00000010, 0b00000001, 0b00000010, 0b00000100, 0b00001000, 0b00000100, 0b00001000, 0b00010000, 0b00000100, 0b00010000, 0b00000100, 0b00000001, 0b00001000, 0b00000001, 0b00001000, 0b00000010, 0b00010000, 0b00000010, 0b00010000, 0b00000001};
static const uint8_t ledDdrD[20]  = {0b00011000, 0b00011000, 0b00110000, 0b00110000, 0b01100000, 0b01100000, 0b11000000, 0b11000000, 0b10100000, 0b10100000, 0b00101000, 0b00101000, 0b01001000, 0b01001000, 0b01010000, 0b01010000, 0b10010000, 0b10010000, 0b10001000, 0b10001000};
static const uint8_t ledPortD[20] = {0b00010000, 0b00001000, 0b00010000, 0b00100000, 0b01000000, 0b00100000, 0b01000000, 0b10000000, 0b00100000, 0b10000000, 0b00100000, 0b00001000, 0b01000000, 0b00001000, 0b01000000, 0b00010000, 0b10000000, 0b00010000, 0b10000000, 0b00001000};
static const uint8_t brightnesses[5][8] = {
    {0, 0, 0, 0, 0, 0, 0, 0}
    , {0, 0, 0, 0, 0, 0, 0, 1}
    , {0, 0, 0, 1, 0, 0, 0, 1}
    , {0, 1, 0, 1, 0, 1, 0, 1}
    , {1, 1, 1, 1, 1, 1, 1, 1}
};


ISR(TIMER1_CAPT_vect)
{
    // 5µs saved by removing the >>3 and using a separate array for PORTB/DDRB (72µs)
    // 58µs saved by removing the need for modulo, which is pretty expensive, and all other pwm-calculations, by using a PWM lookup-table (13.5µs)
    //      who cares about the extra 40 bytes of RAM (out of 2048bytes available)
    // ISR call overhead is an additional 6µs
    // handleButton takes around 21µs (34.5µs in total)

    PORTC |= _BV(PORTC0);//for bebugging timing (ISR takes 65µs-97µs out of 100µs per cycle)

    static uint8_t ledIndex = 0;
    static uint8_t brightnessIndex = 0;
    static uint8_t btnCounter = 0;


    // Left eyelash
    PORTB = 0;
    //if( (brightnessIndex+1) % (16 / (1<<leds[0][ledIndex])) == 0 )    // step+1 is needed for when leds[ledIndex]==0
    if ( brightnesses[leds[0][ledIndex]][brightnessIndex] )
    {
        DDRB  = ledDdrB[ledIndex];
        PORTB = ledPortB[ledIndex];
    }

    // Right eyelash
    PORTD = _BV(PORTD2);//pull-up
    //if( (brightnessIndex+1) % (16 / (1<<leds[1][ledIndex])) == 0 )    // step+1 is needed for when leds[ledIndex]==0
    if ( brightnesses[leds[1][ledIndex]][brightnessIndex] )
    {
        DDRD  = ledDdrD[ledIndex];
        PORTD = ledPortD[ledIndex] | _BV(PORTD2);//-pull-up (|_BV(PD2) is only one single instruction/cycle extra)
    }

    ledIndex++;
    if ( ledIndex == LED_COUNT )
    {
        ledIndex = 0;
        brightnessIndex++;
        if ( brightnessIndex == 8 )
            brightnessIndex = 0;
    }


    btnCounter++;
    if ( btnCounter >= 50 )
    {
        handleButton();
        btnCounter = 0;
    }

    PORTC &= ~_BV(PORTC0);// for debugging timing
    /*
        brightness   dutycycle    step-pattern  16/(1<<brightness)
                    0        0/8       00000000      16
                    1        1/8       00000001      8
                    2        2/8       00010001      4
                    3        4/8       01010101      2
                    4        8/8       11111111      1
    */

}


/*
    ISR(TIMER1_CAPT_vect)
    {
        PORTC |= _BV(PORTC0);//for bebugging timing (ISR takes 34µs out of 257µs per cycle)

        static uint8_t cnt = 0;
        cnt = (cnt+1) % 20;

        // Left eyelash
        PORTB = 0;
        if( leds[0][cnt] )
        {
                DDRB  = ledDdr[cnt]>>3;
                PORTB = ledPort[cnt]>>3;
        }

        // Right eyelash
        PORTD = _BV(PORTD2);//pull-up
        if( leds[1][cnt] )
        {
                DDRD  = ledDdr[cnt];
                PORTD = ledPort[cnt] | _BV(PORTD2);//-pull-up
        }

        if( cnt==0 )
        {
                handleButton();
        }

        PORTC &= ~_BV(PORTC0);// for debugging timing
    }
*/

EMPTY_INTERRUPT(WDT_vect);
EMPTY_INTERRUPT(INT0_vect);
EMPTY_INTERRUPT(INT1_vect);

